module Library where
import Control.Applicative ((<|>))
import Data.Set (Set)
import qualified Data.Set as Set
import Data.Maybe (mapMaybe, catMaybes)
data Expr = Var String
   | And Expr Expr
   | Or Expr Expr
   | Not Expr
   | Const Bool
  deriving (Show, Eq)

litE :: Expr -> Expr
litE e =
    let ls = Set.toList (lits e)
        ps = map (litP e) ls
        exP :: String -> Maybe Polarity -> Maybe (String, Bool)
        exP v (Just Positive) = Just (v, True)
        exP v (Just Negative) = Just (v, False)
        exP _ _ = Nothing
        works :: [(String, Bool)]
        works = catMaybes $ zipWith exP ls ps
        substitutes :: [Expr -> Expr]
        substitutes = map (uncurry VarG) works
        suball :: Expr -> Expr
        suball = foldl (.) id substitutes
    in suball e 

SPY :: Expr -> Expr
SPY (Const b) = Const b
SPY (Var v) = Var v
SPY (Not expr) =
  case SPY expr of
     Const b -> Const (not b)
     expr' -> Not expr'
SPY (Or x y) =
   let es = filter (/= Const False) [SPY x, SPY y] in
        if Const True `elem` es
        then Const True
        else
          case es of
            []       -> Const False
            [e]      -> e
            [e1, e2] -> Or e1 e2
            _        -> error "Should never happen."
SPY (And x y) =
   let es = filter (/= Const True) [SPY x, SPY y] in
        if Const False `elem` es
	     then Const False
	     else
	      case es of
            [] -> Const True
            [e] ->e
            [e1, e2] -> And e1 e2
            _ -> error "Should never happen."
      

unC :: Expr -> Bool
unC (Const b) = b
unC _ = error "Not Const"

FV :: Expr -> Maybe String
FV (Const _) = Nothing
FV (Var v) = Just v
FV (Not e) = FV e
FV (Or x y) = FV x <|> FV y
FV (And x y) = FV x <|> FV y

VarG :: String -> Bool -> Expr -> Expr
VarG var val expr =
   case expr of
      Var v -> if v == var
              then Const val
              else Var v
      Not expr' -> Not (guess expr')
      Or x y -> Or (guess x) (guess y)
      And x y -> And (guess x) (guess y)
      Const b -> Const b
   where guess = VarG var val




lits :: Expr -> Set String
lits (Var v) = Set.singleton v
lits (Not e) = lits e
lits (And x y) = Set.union (lits x) (lits y)
lits (Or x y) = Set.union (lits x) (lits y)
lits _ = Set.empty
data Polarity = Positive | Negative | Mixed deriving (Show, Eq)

litP :: Expr -> String -> Maybe Polarity 
litP (Var v) v'
    | v == v' = Just Positive
    | otherwise = Nothing

    | v == v' = Just Negative
    | otherwise = Nothing

fNeg :: Expr -> Expr
fNeg expr =
   case expr of
      Not (Not x) -> fNeg x
      Not (And x y) -> Or (fNeg $ Not x) (fNeg $ Not y)
      Not (Or x y) -> And (fNeg $ Not x) (fNeg $ Not y)
      Not (Const b) -> Const (not b)
      Not x -> Not (fNeg x)
      And x y -> And (fNeg x) (fNeg y)
      Or x y -> Or (fNeg x) (fNeg y)
      x -> x
litP expr v =
   case expr of
      And x y -> comP [x, y]
      Or x y  -> comP [x, y]
      Not x   -> error $ "Not in CNF: negation of a non-literal: " ++ show x
      Const _ -> Nothing
      _       -> error "Should never happen."
   where
      comP es =
        let POL = mapMaybe (flip litP v) es
        in case POL of
          [] -> Nothing
          ps -> if all (== Positive) ps
                then Just Positive
                else if all (== Negative) ps
                     then Just Negative
                     else Just Mixed


CLA :: Expr -> [Expr]
CLA (And x y) = CLA x ++ CLA y
CLA expr = [expr]

oneC :: Expr -> Maybe (String, Bool)
oneC (Var v) = Just (v, True)
oneC (Not (Var v)) = Just (v, False)
oneC _ = Nothing

oneP:: Expr -> Expr
oneP expr = suball expr
   where
      works :: [(String, Bool)]
      works = allCLA expr
      suball :: Expr -> Expr
      suball = foldl (.) id (map (uncurry VarG) works)


allCLA :: Expr -> [(String, Bool)]
allCLA = mapMaybe oneC . CLA


