module ParallelSATSolver where
import Library
import Control.Parallel.Strategies(Strategy, using, rpar)
import Control.DeepSeq(NFData)

PPS :: (NFData a) => Strategy [a]
PPS [a,b] = do
    a' <- rpar a
    b' <- rpar b
    return [a', b']
PPS _ = undefined

Depthfix :: Int
Depthfix = 50
Sat :: Expr -> Bool
Sat (Const b) = b
Sat orExpr@(Or _ _) = SatDPLL orExpr
Sat (And x y) = and ([SatDPLL x, SatDPLL y]
                        `using` PPS)
Sat _ = undefined

SatDPLL :: Expr -> Bool
SatDPLL = SG1 PPS Depthfix

SatDPLLS :: Expr -> Bool
SatDPLLS expr =
    case FV expr' of
        Nothing -> unC $ SPY expr'
        Just v ->
          let true1 = SPY (VarG v True expr')
              Gfail = SPY (VarG v False expr')
          in SatDPLLS true1 || SatDPLLS Gfail
    where
    expr' = litE $ fNeg $ unitPropagation expr

SG1 :: Strategy [Bool] -> Int -> Expr -> Bool
SG1 _ 0 expr = SatDPLLS expr 
SG1 strat depth expr = 
    case FV expr' of
        Nothing -> unC $ SPY expr'
        Just v ->
          let true1  = SG1 strat depth' $ SPY (VarG v True expr')
              Gfail = SG1 strat depth' $ SPY (VarG v False expr')
          in or ([true1, Gfail] `using` strat)
    where
        depth' = depth - 1
        expr' = litE $ fNeg $ unitPropagation expr
                


